#!/bin/sh

############################################################################
# binary-faker.sh sortof behaves like a couple of binaries used by zogftw.
#
# It reads from stdin and writes to stdout and stderr, but does not touch
# the system. The exit code is supposed to match the one the real binary
# would use for the given input, the rest of the output only matches if
# it matters for zogftw.
#
# Copyright (c) 2014-2022 Fabian Keil <fk@fabiankeil.de>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
############################################################################

show_faked_binaries() {
    grep 'fake_[a-z]*() {$' "$0" | sed -Ee 's@fake_(.*)\(\) {$@\1@'
}

echo_stderr() {
    local message="$*"
    echo "${message}" 1>&2
}

running_as_root() {
    if [ "$RUNNING_AS_ROOT" = 1 -o "$(id -u)" = 0 ]; then
        return 0
    else
        echo_stderr "Not runnign as root!"
        return 1
    fi
}

fake_hostname() {
    case "$*" in
       "-s")
            echo "${FAKED_HOSTNAME}"
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_geli_init() {
    local label metadatabackup \
        passphrase_1 passphrase_2

    label="$2"
    if [ -n "$1" ]; then
        metadatabackup="$1"
    else
        metadatabackup="/var/backups/label_${label}.eli"
    fi

    # XXX: This produces the expected output when called
    #      through zogftw, but does not actually read input
    #      from stdin and produces superfluos newlines when
    #      run interactively.
    echo "Enter new passphrase:"
    read passphrase_1
    echo "Reenter new passphrase:"
    read passphrase_2
    cat<<EOF

Metadata backup can be found in ${metadatabackup} and
can be restored with the following command:

	# geli restore ${metadatabackup} /dev/label/${label}

EOF
    return 0
 }

fake_geli() {
    local faked_geli
    case "$*" in
        'attach /dev/label/fakedpool')
            running_as_root
            return
            ;;
        'attach /dev/label/labelwithoutpool')
            running_as_root || return
            if [ "$DISAPPEARING_LABEL" != 1 ]; then
                for faked_label in $FAKED_LABELS; do
                    if [ "$faked_label" = "label/labelwithoutpool" ]; then
                        return 0
                    fi
                done
            fi
            echo 'geli: Cannot open /dev/label/labelwithoutpool: No such file or directory.' 1>&2
            return 1
            ;;
        "attach /dev/label/newpool"|"attach -k ./newpool.key /dev/label/newpool")
            running_as_root || return
            echo "Enter passphrase:"
            cat > /dev/null
            if [ "$FAKE_WRONG_GELI_KEY" = 1 ]; then
                echo_stderr "geli: Wrong key for label/newpool."
                return 1
            fi
            return 0
            ;;
        "detach label/newpool.eli")
            running_as_root || return
            return 0
            ;;
        "init /dev/label/newpool")
            running_as_root || return
            fake_geli_init "" "newpool"
            ;;
        "init -B ./newpool.eli /dev/label/newpool"|\
        "init -B ./newpool.eli -K ./newpool.key /dev/label/newpool")
            running_as_root || return
            fake_geli_init "./newpool.eli" "newpool"
            ;;
        "init -B /geli/metadata-backups/newpool.eli /dev/label/newpool")
            running_as_root || return
            fake_geli_init "/geli/metadata-backups/newpool.eli" "newpool"
            ;;
        "init -B /geli/metadata-backups/newpool.eli -J - /dev/label/newpool")
            running_as_root || return
            echo "Fake geli init got passphrase '$(cat)' from stdin"
            return 0
            ;;
        "status -s")
            for faked_geli in ${FAKED_GELIS}; do
                echo "${faked_geli}  ACTIVE  ${faked_geli%%.eli}"
            done
            return 0
            ;;
        "status -s label/fakedpool.eli")
            for faked_geli in ${FAKED_GELIS}; do
                if [ "$faked_geli" = "label/fakedpool.eli" ]; then
                    return 0
                fi
            done
            echo_stderr "geli: No such geom: label/fakedpool.eli."
            return 1
            ;;
        "status -s label/labelwithoutpool.eli")
            for faked_geli in ${FAKED_GELIS}; do
                if [ "$faked_geli" = "label/labelwithoutpool.eli" ]; then
                    return 0
                fi
            done
            echo_stderr "geli: No such geom: label/labelwithoutpool.eli."
            return 1
            ;;
        "status -s label/newpool.eli")
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_glabel() {
    local \
        faked_label

    case "$*" in
        "label newpool /dev/null")
            running_as_root
            return
            ;;
       "status -s")
            for faked_label in $FAKED_LABELS; do
                # XXX: Does not fake leading whitespace
                echo "$faked_label N/A abc0"
            done
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_gpg() {
    case "$*" in
        "--encrypt --output ./nonexistantfile")
            # XXX: Should fake stdout output
            return 0
            ;;
        "--encrypt --output ./nonexistantfile_that_cannot_be_created")
            echo_stderr "gpg-fake: Failed to create './nonexistantfile_that_cannot_be_created'"
            return 1
            ;;
        "--decrypt --quiet "*"/tests/config/zogftw/geli/passphrases/newpool.gpg")
            if [ -z "${FAKED_ENCRYPTED_PASSPHRASE}" ]; then
                return 1
            fi
            echo "${FAKED_ENCRYPTED_PASSPHRASE}"
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_mbuffer() {
    case "$*" in
       "-m 100M")
            cat
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_dd() {
    case "$*" in
        'bs=1M if=/dev/zero of=/dev/label/newpool.eli')
            running_as_root
            echo "100+0 records in"
            echo "100+0 records out"
            echo "104857600 bytes transferred in 0.458040 secs (228926597 bytes/sec)"
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_sudo() {
    local \
        original_arguments="$*" \
        binary

    if [ "$1 $2" = "-c -" ]; then
        shift;
        shift;
    fi

    binary=$1
    shift

    case "$binary" in
        dd|geli|glabel|zfs|zpool)
            RUNNING_AS_ROOT=1
            "fake_${binary}" "$@"
            return
            ;;
        false)
            return 1
            ;;
        -v)
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '${original_arguments}'"
            return 1
            ;;
    esac
}

fake_zfs() {
    local \
        stdin_input

    case "$*" in
        "allow sourcepool/ spaces are / awesome/ ")
            echo "---- sourcepool/ spaces are / awesome/  -------------------------------------"
            echo "Local+Descendent permissions:"
            echo "	user $(id -un) create,diff,mount,receive,send"
            return 0
            ;;
        "allow sourcepool/no/sudo/required")
            echo "---- Permissions on sourcepool/no/sudo -------------------------------------"
            echo "Local+Descendent permissions:"
            echo "	user $(id -un) create,diff,mount,receive,send,snapshot"
            return 0
            ;;
        "allow sourcepool/foo/bar")
            echo "---- Permissions on sourcepool/foo/bar -------------------------------------"
            echo "Local+Descendent permissions:"
            echo "	user $(id -un) create,diff,mount,receive,send"
            echo "	user $(id -un)x create,diff,mount,receive,send,snapshot"
            return 0
            ;;
        "create -p -u sourcepool/no/sudo/required"*)
            if [ "$RUNNING_AS_ROOT" = 1 ]; then
                echo_stderr "$0: Tried to create dataset as root"
                return 1
            fi
            return 0
            ;;
        "create -p -u newpool/backup/${FAKED_HOSTNAME}")
            running_as_root
            return
            ;;
        "create -p -u testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die")
            running_as_root
            return
            ;;
        "create -p -u testpool/backup/${FAKED_HOSTNAME}/sourcepool/ spaces are / awesome")
            running_as_root
            return
            ;;
        "get -H -o value de.fabiankeil:zogftw:new_datasets testpool/backup/testhostname")
            if [ -z "${FAKED_ZFS_PROPERTY_VALUE}" ]; then
                FAKED_ZFS_PROPERTY_VALUE=yes
            fi
            echo "${FAKED_ZFS_PROPERTY_VALUE}"
            return
            ;;
        "get -H -s local de.fabiankeil:zogftw:last_snapshot testpool/blafasel/no-snapshot-property-saved")
            return
            ;;
        "get -H -s local de.fabiankeil:zogftw:last_snapshot testpool/blafasel/snapshot-property-saved")
            echo "testpool/blafasel/snapshot-property-saved	de.fabiankeil:zogftw:last_snapshot	testpool/blafasel/snapshot-property-saved@2022-03-05_09:31	local"
            return
            ;;
        "get -H -o value guid"*)
            if [ -z "${FAKED_ZFS_GUID_VALUE}" ]; then
                if [ -z "${FAKED_ZFS_PROPERTY_VALUE}" ]; then
                    echo_stderr "FAKED_ZFS_GUID_VALUE and FAKED_ZFS_PROPERTY_VALUE undefined"
                    return 2
                fi
                FAKED_ZFS_GUID_VALUE="${FAKED_ZFS_PROPERTY_VALUE}"
            fi
            echo "${FAKED_ZFS_GUID_VALUE}"
            return
            ;;
        "get -H -o value"*)
            if [ -z "${FAKED_ZFS_PROPERTY_VALUE}" ]; then
                echo_stderr "bad property list: invalid property [...]"
                return 2
            fi
            echo "${FAKED_ZFS_PROPERTY_VALUE}"
            return
            ;;
        "inherit de.fabiankeil:zogftw:last_snapshot blafasel/die/da")
            return
            ;;
        "list -r -H -o name -s name -t filesystem,volume"*)
            if [ -z "${FAKED_ZFS_DATASETS}" ]; then
                echo_stderr "No fake datasets provided"
                return 2
            fi
            echo "${FAKED_ZFS_DATASETS}"
            return 0
            ;;
        "list -H -r -o name -t filesystem,volume pooltoupdate/backup/testhostname")
            echo "pooltoupdate/backup/testhostname"
            echo "pooltoupdate/backup/testhostname/bla"
            echo "pooltoupdate/backup/testhostname/bla/fasel"
            ;;
        "list -H -o name -s name newpool")
            echo newpool
            return 0
            ;;
        "list -H -o name -s name pooltoupdate")
            echo pooltoupdate
            return 0
            ;;
        "list -H -o name -s name pooltoupdate/backup/testhostname")
            echo pooltoupdate/backup/testhostname
            return 0
            ;;
        "list -H -o name -s name sourcepool/no/sudo/required")
            echo "sourcepool/no/sudo/required"
            return 0
            ;;
        "list -H -o name -s name sourcepool/dataset/does/not/exist")
            echo "cannot open 'sourcepool/dataset/does/not/exist': dataset does not exist"
            return 1
            ;;
        "list -H -o name -s name sourcepool/ spaces are ")
            echo "cannot open 'name sourcepool/ spaces are ': dataset does not exist"
            return 1
            ;;
        "list -H -o name -s name sourcepool/ spaces are / awesome/ ")
            echo "sourcepool/ spaces are / awesome/ "
            return 0
            ;;
        "list -H -o name -s name fakedpool")
            if [ "${FAKED_POOLS}" = "fakedpool" ]; then
                echo_stderr "cannot open 'fakedpool': dataset does not exist"
                return 0
            fi
            return 1
            ;;
        "list -H -o name -s name testpool/backup/${FAKED_HOSTNAME}")
            echo "testpool/backup/${FAKED_HOSTNAME}"
            ;;
        "list -H -o name -s name sourcepool/blafasel/die/da")
            echo "sourcepool/blafasel/die/da"
            ;;
        "list -t snapshot -H -o name -r -d 1 pooltoupdate/backup/testhostname")
            ;;
        "list -t snapshot -H -o name -r -d 1 pooltoupdate/backup/testhostname/bla")
            ;;
        "list -t snapshot -H -o name -r -d 1 pooltoupdate/backup/testhostname/bla/fasel")
            echo "pooltoupdate/backup/testhostname/bla/fasel@2022-03-30_15:34"
            echo "pooltoupdate/backup/testhostname/bla/fasel@2022-03-30_15:35"
            echo "pooltoupdate/backup/testhostname/bla/fasel@2022-03-30_15:36"
            ;;
        "list -t snapshot -H -o name -r -d 1 sourcepool/blafasel/die/da")
            # XXX: Output isn't properly validated by zogftw!
            if [ -z "${FAKED_SOURCE_SNAPSHOT_NAMES}" ]; then
                echo "FAKED_SOURCE_SNAPSHOT_NAMES isn't set" 1>&2
                return 1
            fi
            for name in $FAKED_SOURCE_SNAPSHOT_NAMES; do
                echo "sourcepool/blafasel/die/da@$name"
            done
            ;;
        "list -t snapshot -H -o name -r -d 1 sourcepool/ spaces are / awesome/ ")
            if [ -z "${FAKED_SOURCE_SNAPSHOT_NAMES}" ]; then
                echo "FAKED_SOURCE_SNAPSHOT_NAMES isn't set" 1>&2
                return 1
            fi
            for name in $FAKED_SOURCE_SNAPSHOT_NAMES; do
                echo "sourcepool/ spaces are / awesome/ @$name"
            done
            ;;
        "list -t snapshot -H -o name -r -d 1 testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die/da")
            if [ -z "${FAKED_DESTINATION_SNAPSHOT_NAMES}" ]; then
                echo "FAKED_DESTINATION_SNAPSHOT_NAMES isn't set" 1>&2
                return 1
            fi
            for name in $FAKED_DESTINATION_SNAPSHOT_NAMES; do
                echo "testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die/da@$name"
            done
            ;;
        "list -t snapshot -H -o name -r -d 1 testpool/blafasel")
            if [ -z "${FAKED_SNAPSHOT_NAMES}" ]; then
                return 0
            fi
            for name in $FAKED_SNAPSHOT_NAMES; do
                echo "testpool/blafasel@$name"
            done
            ;;
        "list -t snapshot -H -o name -s name -r -d 1 testpool/chronological")
            if [ "${ZOGFTW_SORTING_SNAPSHOTS_BY_NAME_KEEPS_CHRONOLOGICAL_ORDER}" != 1 ]; then
                echo_stderr "$0: Requested snapshots sorted by name despite the users wishes"
                return 1
            fi
            if [ -z "${FAKED_SNAPSHOT_NAMES}" ]; then
                return 0
            fi
            for name in $FAKED_SNAPSHOT_NAMES; do
                echo "testpool/chronological@$name"
            done | sort
            ;;
        "list -H -o name -s name sourcepool/foo/bar")
            echo "sourcepool/foo/bar"
            ;;
        "list -H -o name -s name /"*"${FAKED_DATASET}")
            if [ -z "${FAKED_DATASET}" ]; then
                echo_stderr "[...]: No such file or directory"
                return 1
            fi
            echo "${FAKED_DATASET}"
            ;;
        "list -t snapshot -H -o name -r -d 1 sourcepool/foo/bar")
            ;;
        "list -H -o name -s name testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die/da")
            if [ ! -z "${FAKED_DESTINATION_DATASET_EXISTS}" ]; then
                echo "testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die/da"
                return 0
            fi
            return 1
            ;;
        "send testpool/backup/${FAKED_HOSTNAME}")
            running_as_root
            return
            ;;
        "set compression=lz4 newpool/backup/${FAKED_HOSTNAME}")
            running_as_root
            return
            ;;
        "receive testpool/backup/${FAKED_HOSTNAME}/sourcepool/ spaces are / awesome/ "|\
        "receive testpool/backup/${FAKED_HOSTNAME}/sourcepool/blafasel/die/da")
            running_as_root || return

            stdin_input="$(cat)"
            if [ -n "$stdin_input" ]; then
                echo "Fake sudo: 'zfs received': $stdin_input"
                return 0
            else
                echo "Fake sudo: Did not 'zfs receive' anything"
                return 1
            fi
            ;;
        "send sourcepool/ spaces are / awesome/ @2014-08-09_13:18")
            running_as_root
            echo "Fake stream package for sourcepool/ spaces are / awesome/ @2014-08-09_13:18"
            return
            ;;
        "send sourcepool/blafasel/die/da@2014-08-09_13:18")
            running_as_root || return
            echo "Fake stream package for sourcepool/blafasel/die/da@2014-08-09_13:18"
            return 0
            ;;
        "send -I @2014-08-09_13:18 sourcepool/blafasel/die/da@2014-08-09_13:20")
            running_as_root || return
            echo "Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20"
            return 0
            ;;
        "send -I @2014-08-09_13:19 sourcepool/blafasel/die/da@2014-08-09_13:20")
            running_as_root || return
            echo "Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:19. End: @2014-08-09_13:20"
            return 0
            ;;
        "send -I @2013-08-09_13:18 sourcepool/blafasel/die/da@2014-08-09_13:20")
            running_as_root || return
            echo_stderr "WARNING: could not send sourcepool/blafasel/die/da@2014-08-09_13:18:"
            echo_stderr "incremental source (sourcepool/blafasel/die/da@2013-08-09_13:18) does not exist"
            return 1
            ;;
        "send -t invalid-token")
            echo_stderr "cannot resume send: resume token is corrupt (invalid format)"
            return 1
            ;;
        "send -t 1-f504d3a71-d8-789c636064000310a500c4ec50360710e72765a52697303054f042d460c8a7a515a7968064b8e0f26c48f2499525a9c5403a83e3e843262cfa4bf2d34b3353181856345e5bcd7f95e59207923c27583e2f3137154827e665eb67e4e7a6eaa765eb2796a664e6eb97e5172565163b1819189aea1a1aea1a58c61b1a59199a31c00100cb3e2176")
            echo_stderr "Faking resumed send operation"
            echo "Fake stream package based on token"
            return 0
            ;;
        "send -P -n -t flawed-token-1")
            echo_stderr "cannot resume send: resume token is corrupt (invalid format)"
            return 255
            ;;
        "send -P -n -t valid-token-1")
            cat<<EOF
resume token contents:
nvlist version: 0
	object = 0xd87
	offset = 0xa0000
	bytes = 0x21ec58086
	toguid = 0xd20d450afbd618a8
	toname = sourcepool/blafasel/die/da@2014-08-09_13:20
full	sourcepool/blafasel/die/da@2014-08-09_13:20	54213913992
EOF
            return 1
            ;;
        "send -t valid-token-1")
            echo "Fake zfs send stream resumed using valid-token-1"
            ;;
        "set de.fabiankeil:zogftw:last_snapshot=pooltoupdate/backup/testhostname/bla/fasel@2022-03-30_15:36 pooltoupdate/backup/testhostname/bla/fasel")
            ;;
        "snapshot sourcepool/foo/bar@"*)
            running_as_root
            return
            ;;
        "snapshot sourcepool/ spaces are / awesome/ @"*)
            running_as_root
            return
            ;;
        "snapshot sourcepool/no/sudo/required@"*)
            if [ "$RUNNING_AS_ROOT" = 1 ]; then
                echo_stderr "$0: Tried to snapshot as root"
                return 1
            fi
            return 0
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

fake_zpool() {
    local \
        faked_zpool

    case "$*" in
        "export fakedpool"|'export newpool'|'export -f newpool')
            running_as_root
            return
            ;;
        'create newpool /dev/label/newpool.eli')
            running_as_root
            return
            ;;
        "import -d /dev/label labelwithoutpool")
            running_as_root || return
            echo_stderr "cannot import 'labelwithoutpool': no such pool available"
            return 1;
            ;;
        "import -d /dev/label fakedpool")
            running_as_root
            return
            ;;
        "list -H -o name")
            for faked_zpool in ${FAKED_ZPOOLS}; do
                echo "${faked_zpool}"
            done
            ;;
        "list -H -o name fakedpool")
            for faked_zpool in ${FAKED_ZPOOLS}; do
                if [ "$faked_zpool" = "fakedpool" ]; then
                    echo "${faked_zpool}"
                    return 0
                fi
            done
            echo_stderr "cannot open 'fakedpool': no such pool"
            return 1
            ;;
        "list -H -o name nonexistantpool")
            echo "cannot open 'nonexistantpool': no such pool"
            return 1
            ;;
        "list -H -o name testpool")
            echo "testpool"
            ;;
        "list -H -o size testpool")
            if [ -n "${FAKED_ZPOOL_SIZE}" ]; then
                echo "${FAKED_ZPOOL_SIZE}"
            else
                echo "100G"
            fi
            ;;
        "list -H -p -o size testpool")
            if [ -n "${FAKED_ZPOOL_SIZE}" ]; then
                echo "${FAKED_ZPOOL_SIZE}"
            else
                echo "107374182400" # 100G
            fi
            ;;
        "status fakedpool")
            for faked_zpool in ${FAKED_ZPOOLS}; do
                if [ "$faked_zpool" = "fakedpool" ]; then
                    echo "  pool: ${faked_zpool}"
                    echo " state: ONLINE"
                    echo "  scan: scrub repaired 0 in 0h9m with 0 errors on Tue Jul  8 10:46:19 2014"
                    echo "config:"
                    echo ""
                    echo "	NAME                   STATE     READ WRITE CKSUM"
                    echo "	${faked_zpool}              ONLINE       0     0     0"
                    echo "	  label/${faked_zpool}.eli  ONLINE       0     0     0"
                    echo ""
                    echo "errors: No known data errors"
                    return 0
                fi
            done
            echo_stderr "cannot open 'fakedpool': no such pool"
            return 1
            ;;
        *)
            echo_stderr "$0: Unexpected input: '$*'"
            return 1
            ;;
    esac
}

main() {
    local faked_binary
    faked_binary=${0##*/}

    : "${FAKED_HOSTNAME=testhostname}"
    : "${RUNNING_AS_ROOT=0}"

    if [ "$1" = "show-fakes" ]; then
            show_faked_binaries
            return 0
    fi
    case "${faked_binary}" in
        hostname|geli|glabel|gpg|mbuffer|sudo|zfs|zpool)
            "fake_${faked_binary}" "$@"
            ;;
        *)
            echo_stderr "$0: Unexpected binary mode. Input: '$*'"
            return 1
            ;;
    esac
}

main "$@"
