#!/usr/bin/env atf-sh
############################################################################
# Copyright (c) 2014-2022 Fabian Keil <fk@fabiankeil.de>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
############################################################################

get_test_cases() {
    grep '.*_body() {$' "$0" | sed -e 's@_body() {$@@'
}

# The registration has to happen before the test_case bodies
# are declared, otherwise they the bodies will be overwritten
register_test_cases() {
    local \
        test_case

    for test_case in $(get_test_cases); do
        atf_test_case "${test_case}"
    done
}
register_test_cases

atf_init_test_cases() {
    local \
        test_case

    export ZOGFTW_CONFIG_FILE=''
    ZOGFTW="$(atf_get_srcdir)/../zogftw"

    for test_case in $(get_test_cases); do
        atf_add_test_case "${test_case}"
    done
}

# Makes sure real ZFS pools or datasets aren't touched.
enable_pseudo_executables() {
    local \
        pseudo_executable_path="$(atf_get_srcdir)/mockery" \
        executable

    for executable in hostname geli glabel gpg mbuffer sudo zfs zpool; do
        if [ ! -x "${pseudo_executable_path}/$executable" ]; then
            atf_fail "Missing  executable '$executable' in ${pseudo_executable_path}"
        fi
    done
    export PATH="${pseudo_executable_path}:$PATH"
}

# Source zogftw and abort testing if it fails
source_zogftw() {
    # For zogftw_get_function_prototypes() to work,
    # zogftw has to be in the PATH.
    export PATH="${PATH}:$(atf_get_srcdir)/../"

    # Some shells don't pass arguments to sourced scripts,
    # manually setting $1 to "source" works around this.
    set -- source
    . "${ZOGFTW}" source || atf_fail "Failed to source zogftw"
}

detect_environment_pollution_body() {
    environment_seems_clean || atf_fail "Environment is polluted"
}

environment_seems_clean() {
    local \
        variable content

    for variable in $("${ZOGFTW}" cmd zogftw_get_config_variables); do
        content="$(eval echo "\$${variable}")"
        if [ -n "$content" ]; then
            echo "Environment variable ${variable} is set to '${content}'."
            return 1
        fi
    done

    return 0
}

echo_supports_n_body() {
    echo "Expecting a newline"     | atf_check -s exit:0 -o "inline:Expecting a newline\n" -e empty cat
    echo -n "Expecting no newline" | atf_check -s exit:0 -o "inline:Expecting no newline"  -e empty cat
}

# Test case fails if there is a zogftw function that doesn't
# have at least one test case and isn't a hook. Technically
# hooks should be tested, too, but currently the test coverage
# is known to be spotty. Also, hooks probably don't need their
# own test case anyway.
no_untested_zogftw_functions_body() {
    local \
        zogftw_function untested_functions

    get_test_cases | grep ^zogftw | sort > zogftw_function_test_cases

    for zogftw_function in $("${ZOGFTW}" cmd zogftw_get_function_prototypes \
        | grep -v _hook | cut -d "(" -f 1); do
        if ! grep -Eq "^${zogftw_function}(_success|_failures|\$)" zogftw_function_test_cases; then
            untested_functions="$untested_functions ${zogftw_function}"
        fi
    done

    if [ -n "${untested_functions}" ]; then
        atf_fail "Missing tests for:${untested_functions}"
    fi
}

zogftw_change_space_escaping_body() {
    atf_check -s exit:0 -o "inline:\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping ""
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a\ b\ c\ d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a\ b\ c%20d"
    atf_check -s exit:0 -o "inline:a%20b%20c d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a%20b\ c d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a\ b\ c\ \ d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a\ b\ c\ %20d"
    atf_check -s exit:0 -o "inline:a%20b%20%20c%20%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "a\ b\ \ c\ \ \ d"
    atf_check -s exit:0 -o "inline:%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "\ "
    atf_check -s exit:0 -o "inline:%20%20%20%20%20%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_change_space_escaping "\ \ \ \ \ \ "
}

zogftw_clear_device_body() {
    local \
        device_name

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    device_name="/dev/label/newpool.eli"

    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_clear_device_newpool_eli" \
        -e empty "${ZOGFTW}" cmd zogftw_clear_device "${device_name}"

    atf_check -s exit:1 \
        -o empty -e 'inline:zogftw_clear_device(): No device to clear provided\n' \
        "${ZOGFTW}" cmd zogftw_clear_device
}

zogftw_config_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_config" \
        -e empty "${ZOGFTW}" cmd zogftw_config
}

# zogftw_cmd() is already indirectly tested by most of the other
# test cases anyway, but test cases are cheap.
zogftw_cmd_body() {
    atf_check -s exit:0 -o "inline:bla\n" \
        -e empty "${ZOGFTW}" cmd zogftw_cmd echo bla
    # Test passing of positional arguments
    atf_check -s exit:0 -o "inline:bla fasel die da trallala\n" \
        -e empty "${ZOGFTW}" cmd zogftw_cmd eval " echo bla fasel die da" trallala
    atf_check -s exit:0 -o "inline:bla fasel die da\n" \
        -e empty "${ZOGFTW}" cmd zogftw_cmd sh -c "echo bla fasel die da" argument-that-is-expected-to-get-lost
}

zogftw_create_encrypted_geli_passphrase_file_gpg_failure_body() {
    local \
        passphrase_file

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    create_stty_less_zogftw_copy

    passphrase_file=./nonexistantfile_that_cannot_be_created
    printf "Passphrase\nPassphrase\n" | \
        atf_check -s exit:1 -o "inline:Enter geli passphrase: \nReenter geli passphrase: \nWriting passphrase to ${passphrase_file}\n" \
        -e "inline:gpg-fake: Failed to create '${passphrase_file}'\n" \
        ./zogftw cmd zogftw_create_encrypted_geli_passphrase_file "${passphrase_file}"
}

zogftw_create_encrypted_geli_passphrase_file_invalid_arguments_body() {
    local \
        passphrase_file

    atf_check -s exit:1 -o empty -e "inline:No passphrase file provided\n" \
        "${ZOGFTW}" cmd zogftw_create_encrypted_geli_passphrase_file

    passphrase_file=./existing_file
    touch "${passphrase_file}"
    atf_check -s exit:1 -o empty -e "inline:Passphrase file '${passphrase_file}' already exists\n" \
        "${ZOGFTW}" cmd zogftw_create_encrypted_geli_passphrase_file "${passphrase_file}"
}

zogftw_create_encrypted_geli_passphrase_file_passphrase_mismatch_body() {
    local \
        passphrase_file

    passphrase_file=./existing_file
    touch "${passphrase_file}"

    create_stty_less_zogftw_copy

    passphrase_file=./nonexistantfile
    printf "Passphrase\npassphrase\n" | \
        atf_check -s exit:1 -o "inline:Enter geli passphrase: \nReenter geli passphrase: \n" \
        -e "inline:Passphrases do not match!\n" \
        ./zogftw cmd zogftw_create_encrypted_geli_passphrase_file "${passphrase_file}"
}

zogftw_create_encrypted_geli_passphrase_file_success_body() {
    local \
        passphrase_file

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    create_stty_less_zogftw_copy

    passphrase_file=./nonexistantfile_that_cannot_be_created
    printf "Passphrase\nPassphrase\n" | \
        atf_check -s exit:1 -o "inline:Enter geli passphrase: \nReenter geli passphrase: \nWriting passphrase to ${passphrase_file}\n" \
        -e "inline:gpg-fake: Failed to create '${passphrase_file}'\n" \
        ./zogftw cmd zogftw_create_encrypted_geli_passphrase_file "${passphrase_file}"
}

zogftw_create_failures_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e 'inline:zogftw_create(): No such device: /dev/doesnotexist0\n' \
        "${ZOGFTW}" create newpool /dev/doesnotexist0
    atf_check -s exit:1 -o empty \
        -e "inline:Label 's p a c e s' contains spaces which is currently unsupported\n" \
        "${ZOGFTW}" create "s p a c e s" /dev/null
}

zogftw_create_success_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    printf "Passphrase\nPassphrase\n"| atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_create_newpool" \
        -e empty "${ZOGFTW}" create newpool /dev/null
}

zogftw_create_missing_parent_datasets_body() {
    local \
        parent_dataset dataset

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_create_missing_parent_datasets /dataset-lacking-parent

    parent_dataset="testpool/backup/$(hostname -s)/sourcepool/blafasel/die"
    dataset="${parent_dataset}/da"
    atf_check -s exit:0 \
        -o "inline:Creating ${parent_dataset} as parent for ${dataset}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_create_missing_parent_datasets "${dataset}"

    parent_dataset="sourcepool/no/sudo/required"
    dataset="${parent_dataset}/but-it-does-not-matter-as-the-parent-exists"
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_create_missing_parent_datasets "${dataset}"

    parent_dataset="sourcepool/no/sudo/required/and"
    dataset="${parent_dataset}/it-matters"
    atf_check -s exit:0 \
        -o "inline:Creating ${parent_dataset} as parent for ${dataset}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_create_missing_parent_datasets "${dataset}"
}

zogftw_dataset_does_exist_body() {
    local \
        existant_dataset non_existant_dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    for existant_dataset in sourcepool/foo/bar \
        sourcepool/blafasel/die/da sourcepool/no/sudo/required; do
        atf_check -s exit:0 -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_dataset_does_exist "${existant_dataset}"
    done

    non_existant_dataset=sourcepool/dataset/does/not/exist
    atf_check -s exit:1 -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_dataset_does_exist "${non_existant_dataset}"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_dataset_does_exist(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_dataset_does_exist
}

zogftw_dataset_has_been_specified_by_path_body() {
    local \
        test_directory

    for test_directory in "bla/fasel" "bla / fasel /  with spaces"; do
        mkdir -p "${test_directory}" || atf_fail "Failed to create test directory ${test_directory}"
        atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_been_specified_by_path "${test_directory}"
    done
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_been_specified_by_path /nonexisting/bla/fasel
    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_been_specified_by_path nonexisting/bla/fasel
}

zogftw_dataset_has_snapshots_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_snapshots sourcepool/foo/bar

    for dataset in sourcepool/blafasel/die/da "sourcepool/ spaces are / awesome/ "; do
        export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18"
        atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_snapshots "${dataset}"

        export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
        atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_dataset_has_snapshots "${dataset}"
    done
}

zogftw_delete_last_snapshot_property_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 \
        -o empty -e empty "${ZOGFTW}" cmd zogftw_delete_last_snapshot_property blafasel/die/da
}

zogftw_echo_list_members_with_attribute_body() {
    atf_check -s exit:0 \
        -o "inline:a bla\nb bla\nc bla\nd bla\n" \
        -e empty "${ZOGFTW}" cmd zogftw_echo_list_members_with_attribute "bla" a b c d
    atf_check -s exit:0 \
        -o "inline:d bla\nc bla\nb bla\na bla\n" \
        -e empty "${ZOGFTW}" cmd zogftw_echo_list_members_with_attribute "bla" d c b a
    atf_check -s exit:0 \
        -o "inline:a b bla\nc bla\nd e bla\nf bla\n" \
        -e empty "${ZOGFTW}" cmd zogftw_echo_list_members_with_attribute "bla" "a b" c "d e" f
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_echo_list_members_with_attribute "bla"
}

zogftw_escape_spaces_body() {
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_escape_spaces: No string provided\n" \
        "${ZOGFTW}" cmd zogftw_escape_spaces ""
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a b c d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a b c%20d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a%20b%20c d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a b c  d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a b c %20d"
    atf_check -s exit:0 -o "inline:a%20b%20%20c%20%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "a b  c   d"
    atf_check -s exit:0 -o "inline:%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces " "
    atf_check -s exit:0 -o "inline:%20%20%20%20%20%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_escape_spaces "      "
}

zogftw_export_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:Can't export blafasel: No such pool\n" \
        "${ZOGFTW}" export blafasel
    export FAKED_POOLS=fakedpool
    atf_check -s exit:0 \
        -o "inline:Exporting fakedpool\nlabel/fakedpool.eli is already gone!\n" \
        -e empty \
        "${ZOGFTW}" export fakedpool
    export FAKED_POOLS=newpool
    export FAKED_GELIS=label/newpool.eli
    atf_check -s exit:0 \
        -o "inline:Exporting newpool\n" -e empty \
        "${ZOGFTW}" export newpool
}

zogftw_export_external_zpool_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:Can't export blafasel: No such pool\n" \
        "${ZOGFTW}" cmd zogftw_export_external_zpool blafasel
    export FAKED_POOLS=fakedpool
    atf_check -s exit:0 \
        -o "inline:Exporting fakedpool\nlabel/fakedpool.eli is already gone!\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_export_external_zpool fakedpool
    export FAKED_POOLS=newpool
    export FAKED_GELIS=label/newpool.eli
    atf_check -s exit:0 \
        -o "inline:Exporting newpool\n" -e empty \
        "${ZOGFTW}" cmd zogftw_export_external_zpool newpool
    atf_check -s exit:0 \
        -o "inline:Exporting newpool\n" -e empty \
        "${ZOGFTW}" cmd zogftw_export_external_zpool -f newpool
}

zogftw_foreach_zpool_failures_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:No zpool subcommand specified\n" \
        "${ZOGFTW}" cmd zogftw_foreach_zpool
    atf_check -s exit:1 \
        -o empty \
        -e "inline:No imported zpools managed by zogftw detected\n" \
        "${ZOGFTW}" cmd zogftw_foreach_zpool status
    export FAKED_GELIS="ada5s23.eli ada10s3.eli"
    atf_check -s exit:1 \
        -o empty \
        -e "inline:No imported zpools managed by zogftw detected\n" \
        "${ZOGFTW}" cmd zogftw_foreach_zpool status
    export FAKED_GELIS="ada5s23.eli ada10s3.eli label/fakedpool.eli"
    atf_check -s exit:1 \
        -o empty \
        -e "inline:No imported zpools managed by zogftw detected\n" \
        "${ZOGFTW}" cmd zogftw_foreach_zpool status
}

zogftw_foreach_zpool_success_body() {
    local \
        expected_output_file

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    expected_output_file="$(atf_get_srcdir)/expected_output/zogftw_zpool_status"

    export FAKED_GELIS="ada5s23.eli label/fakedpool.eli ada10s3.eli"
    export FAKED_ZPOOLS="fakedpool"
    atf_check -s exit:0 -o "file:${expected_output_file}" -e empty \
        "${ZOGFTW}" cmd zogftw_foreach_zpool status

    # From zogftw's point of view there appear to be three different pools
    # with the same name.
    export FAKED_GELIS="label/fakedpool.eli label/fakedpool.eli label/fakedpool.eli"
    for i in 1 2 3; do
        cat "${expected_output_file}" >> triplicated-output
    done
    atf_check -s exit:0 -o "file:triplicated-output" -e empty \
        "${ZOGFTW}" cmd zogftw_foreach_zpool status

    export FAKED_GELIS="label/fakedpool.eli"
    atf_check -s exit:0 -o "inline:Executing: zpool export  fakedpool \n" -e empty \
        "${ZOGFTW}" cmd zogftw_foreach_zpool export

}

zogftw_fyi_body() {
    atf_check -s exit:0 -o "inline:Blafasel\n" -e empty "${ZOGFTW}" cmd zogftw_fyi "Blafasel"
    atf_check -s exit:0 -o "inline:\n"         -e empty "${ZOGFTW}" cmd zogftw_fyi ""
}

zogftw_geli_attach_failure_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKE_WRONG_GELI_KEY=1
    export ZOGFTW_GELI_KEYFILE_DIR=.
    printf "Passphrase\nPassphrase\n" | atf_check -s exit:1 \
        -o "inline:Enter passphrase:\n" \
        -e "inline:geli: Wrong key for label/newpool.\n" \
        "${ZOGFTW}" cmd zogftw_geli_attach newpool /dev/null
    touch ./newpool.key
    printf "Passphrase\nPassphrase\n" | atf_check -s exit:1 \
        -o "inline:Using geli keyfile ./newpool.key\nEnter passphrase:\n" \
        -e "inline:geli: Wrong key for label/newpool.\n" \
        "${ZOGFTW}" cmd zogftw_geli_attach newpool /dev/null
}

zogftw_geli_attach_success_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_GELI_KEYFILE_DIR=.
    printf "Passphrase\nPassphrase\n" | atf_check -s exit:0 \
        -o "inline:Enter passphrase:\n'newpool' attached\n" \
        -e empty "${ZOGFTW}" cmd zogftw_geli_attach newpool /dev/null
    touch ./newpool.key
    printf "Passphrase\nPassphrase\n" | atf_check -s exit:0 \
        -o "inline:Using geli keyfile ./newpool.key\nEnter passphrase:\n'newpool' attached\n" \
        -e empty "${ZOGFTW}" cmd zogftw_geli_attach newpool /dev/null
}

zogftw_geli_geom_exists_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e 'inline:zogftw_geli_geom_exists: No geli geom to check provided\n' \
        "${ZOGFTW}" cmd zogftw_geli_geom_exists
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_geli_geom_exists label/labelwithoutpool.eli
    export FAKED_GELIS=label/labelwithoutpool.eli
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_geli_geom_exists label/labelwithoutpool.eli
}

zogftw_geli_initialize_success_with_keyfile_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_GELI_KEYFILE_DIR=.
    export ZOGFTW_GELI_METADATA_BACKUP_DIR=.
    touch newpool.key

    cat > ./expected_output <<EOF
Using geli keyfile ./newpool.key
Initializing /dev/label/newpool with: -B ./newpool.eli -K ./newpool.key
Enter new passphrase:
Reenter new passphrase:

Metadata backup can be found in ./newpool.eli and
can be restored with the following command:

	# geli restore ./newpool.eli /dev/label/newpool

EOF
    atf_check -s exit:0 -o "file:./expected_output" -e empty \
        "${ZOGFTW}" cmd zogftw_geli_initialize /dev/label/newpool
}

zogftw_geli_initialize_success_without_keyfile_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    cat > ./expected_output <<EOF
No geli keyfile found at /geli/keyfiles/newpool.key. Not using any.
Initializing /dev/label/newpool with: -B /geli/metadata-backups/newpool.eli
Enter new passphrase:
Reenter new passphrase:

Metadata backup can be found in /geli/metadata-backups/newpool.eli and
can be restored with the following command:

	# geli restore /geli/metadata-backups/newpool.eli /dev/label/newpool

EOF
    atf_check -s exit:0 -o "file:./expected_output" \
        -e empty "${ZOGFTW}" cmd zogftw_geli_initialize /dev/label/newpool
}

zogftw_geli_initialize_success_not_saving_metadata_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_GELI_METADATA_BACKUP_DIR=""
    cat > ./expected_output <<EOF
No geli keyfile found at /geli/keyfiles/newpool.key. Not using any.
Initializing /dev/label/newpool with: 
Enter new passphrase:
Reenter new passphrase:

Metadata backup can be found in /var/backups/label_newpool.eli and
can be restored with the following command:

	# geli restore /var/backups/label_newpool.eli /dev/label/newpool

EOF
    atf_check -s exit:0 -o "file:./expected_output" -e empty \
        "${ZOGFTW}" cmd zogftw_geli_initialize /dev/label/newpool
}

zogftw_geli_initialize_using_existing_gpg_passphrase_file_body() {
    local \
        expected_geli_passphrase='Super-secret geli passphrase' \
        gnupg_passphrase="gnupg passphrase"

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_GELI_PASSPHRASE_DIR="$(atf_get_srcdir)/config/zogftw/geli/passphrases"

    cat > ./expected_output <<EOF
No geli keyfile found at /geli/keyfiles/newpool.key. Not using any.
Getting the passphrase for /dev/label/newpool with: zogftw_get_passphrase
Initializing /dev/label/newpool with: -B /geli/metadata-backups/newpool.eli -J -
Fake geli init got passphrase '${expected_geli_passphrase}' from stdin
EOF

    export FAKED_ENCRYPTED_PASSPHRASE="${expected_geli_passphrase}"
    echo "${gnupg_passphrase}" | atf_check -s exit:0 \
        -o "file:./expected_output" -e empty \
        "${ZOGFTW}" cmd zogftw_geli_initialize /dev/label/newpool
}

zogftw_geli_initialize_failure_due_to_existing_metadata_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_GELI_METADATA_BACKUP_DIR=.
    touch newpool.eli
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_geli_initialize(): Metadata backup file: './newpool.eli' already exists\n" \
        "${ZOGFTW}" cmd zogftw_geli_initialize /dev/label/newpool
}

zogftw_get_all_snapshots_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_all_snapshots(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots

    dataset=testpool/blafasel
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3"
    atf_check -s exit:0 \
        -o "inline:${dataset}@1\n${dataset}@2\n${dataset}@3\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die a"
    atf_check -s exit:0 \
        -o "inline:${dataset}@bla\n${dataset}@fasel\n${dataset}@die\n${dataset}@a\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"

    export ZOGFTW_SORTING_SNAPSHOTS_BY_NAME_KEEPS_CHRONOLOGICAL_ORDER=1
    dataset=testpool/chronological
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3"
    atf_check -s exit:0 -o "inline:${dataset}@1\n${dataset}@2\n${dataset}@3\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel a"
    atf_check -s exit:0 -o "inline:${dataset}@a\n${dataset}@bla\n${dataset}@fasel\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_all_snapshots "${dataset}"
}

zogftw_get_config_variables_body() {
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_get_config_variables" \
        -e empty "${ZOGFTW}" cmd zogftw_get_config_variables
}

zogftw_get_dataset_from_path_body() {
    local \
        path
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_dataset_from_path nonexistant

    if [ "$(id -u)" != 0 ]; then
        path="blafasel"
        mkdir "${path}"
        chmod -x "${path}"
        atf_check -s exit:1 -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_get_dataset_from_path "${path}"
    fi

    path="path/to/potential/dataset"
    mkdir -p "${path}"
    atf_check -s exit:1 -o empty -e "inline:[...]: No such file or directory\n" \
        "${ZOGFTW}" cmd zogftw_get_dataset_from_path "${path}"

    export FAKED_DATASET="${path}"
    atf_check -s exit:0 -o "inline:${FAKED_DATASET}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_dataset_from_path "${path}"
}

zogftw_get_exportable_zpools_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_exportable_zpools

    export FAKED_GELIS="label/fakedpool.eli"
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_exportable_zpools

    export FAKED_ZPOOLS=fakedpool
    atf_check -s exit:0 -o "inline:${FAKED_ZPOOLS}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_exportable_zpools

    export FAKED_GELIS="label/labelwithoutpool.eli label/fakedpool.eli label/labelwithoutpool1.eli"
    atf_check -s exit:0 -o "inline:${FAKED_ZPOOLS}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_exportable_zpools

    # From zogftw's point of view, we are faking three different pools ...
    export FAKED_GELIS="label/fakedpool.eli label/fakedpool.eli label/fakedpool.eli"
    export FAKED_ZPOOLS="fakedpool fakedpool fakedpool"
    atf_check -s exit:0 -o "inline:${FAKED_ZPOOLS}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_exportable_zpools
}

zogftw_get_first_snapshot_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_first_snapshot(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot

    dataset=testpool/blafasel
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:${dataset}@1\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die a"
    atf_check -s exit:0 -o "inline:${dataset}@bla\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"

    export ZOGFTW_SORTING_SNAPSHOTS_BY_NAME_KEEPS_CHRONOLOGICAL_ORDER=1
    dataset=testpool/chronological
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:${dataset}@1\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die a"
    atf_check -s exit:0 -o "inline:${dataset}@a\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_first_snapshot "${dataset}"
}

zogftw_get_function_prototypes_body() {
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_get_function_prototypes" \
        -e empty "${ZOGFTW}" cmd zogftw_get_function_prototypes
}

zogftw_get_fyi_prefix_body() {
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_get_fyi_prefix
}

zogftw_get_asciidoc_body() {
    atf_check -s exit:0 -o "file:$(atf_get_srcdir)/expected_output/zogftw_get_asciidoc" \
        -e empty "${ZOGFTW}" cmd zogftw_get_asciidoc
}

zogftw_get_last_snapshot_from_property_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_last_snapshot_from_property(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_from_property

    dataset=testpool/blafasel/no-snapshot-property-saved
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_from_property "${dataset}"

    dataset=testpool/blafasel/snapshot-property-saved
    atf_check -s exit:0 -o "inline:testpool/blafasel/snapshot-property-saved@2022-03-05_09:31\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_from_property "${dataset}"
}


zogftw_get_last_snapshot_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_last_snapshot(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot

    dataset=testpool/blafasel
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:${dataset}@4\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die da"
    atf_check -s exit:0 -o "inline:${dataset}@da\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"

    export ZOGFTW_SORTING_SNAPSHOTS_BY_NAME_KEEPS_CHRONOLOGICAL_ORDER=1
    dataset=testpool/chronological
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:${dataset}@only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:${dataset}@4\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die da"
    atf_check -s exit:0 -o "inline:${dataset}@fasel\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot "${dataset}"
}

zogftw_get_last_snapshot_name_body() {
    local \
        dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_last_snapshot_name(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name

    dataset=testpool/blafasel
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:4\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die da"
    atf_check -s exit:0 -o "inline:da\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"

    export ZOGFTW_SORTING_SNAPSHOTS_BY_NAME_KEEPS_CHRONOLOGICAL_ORDER=1
    dataset=testpool/chronological
    export FAKED_SNAPSHOT_NAMES=""
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="only-one-snapshot"
    atf_check -s exit:0 -o "inline:only-one-snapshot\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="1 2 3 4"
    atf_check -s exit:0 -o "inline:4\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
    export FAKED_SNAPSHOT_NAMES="bla fasel die da"
    atf_check -s exit:0 -o "inline:fasel\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_last_snapshot_name "${dataset}"
}

zogftw_get_new_snapshot_name_body() {
    if [ "$(date +%S)" -gt 58 ]; then
        # Prevent the test from failing if the minute
        # (%M) changes while the test is running.
        sleep 2;
    fi
    atf_check -s exit:0 \
        -o "inline:$(date +%Y-%m-%d_%H:%M)\n" \
        -e empty "${ZOGFTW}" cmd zogftw_get_new_snapshot_name
}

zogftw_get_passphrase_body() {
    local \
        passphrase_file \
        expected_geli_passphrase='Super-secret geli passphrase\n' \
        gnupg_passphrase="gnupg passphrase"

    atf_require_prog gpg2

    # To simplify the test, the test file is encrypted
    # with a symmetric cipher using a passphrase
    passphrase_file="$(atf_get_srcdir)"/config/zogftw/geli/passphrases/newpool.gpg

    # --quiet is the default, the rest is required for the test case to work.
    export ZOGFTW_GPG_DECRYPT_FLAGS="--quiet --no-tty --batch --passphrase-fd 0 --homedir $(atf_get_srcdir)/config/gnupg"

    echo "${gnupg_passphrase}" | atf_check \
        -o "inline:${expected_geli_passphrase}" -e 'ignore' \
        gpg --decrypt ${ZOGFTW_GPG_DECRYPT_FLAGS} "${passphrase_file}"

    echo "${gnupg_passphrase}" | atf_check \
        -o "inline:${expected_geli_passphrase}" -e 'ignore' \
        "${ZOGFTW}" cmd zogftw_get_passphrase "${passphrase_file}"

    echo "Not the right passphrase" | atf_check \
        -o "ignore" -e 'ignore' -s 'exit:2' \
        "${ZOGFTW}" cmd zogftw_get_passphrase "${passphrase_file}"

    # More recent gpg versions leave an empty file we don't care about
    rm "$(atf_get_srcdir)/config/gnupg/.gpg-v21-migrated" || true
}

zogftw_get_sorted_src_dataset_information_body() {
    local \
        required_dataset_info \
        optional_dataset_info \
        exteral_dataset_info

    atf_check -o "empty" -e 'empty' -s 'exit:0' \
        "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    export ZOGFTW_REQUIRED_SRC_DATASETS="tank/home/zogftw/blafasel tank/home/zogftw"
    required_dataset_info="tank/home/zogftw required\ntank/home/zogftw/blafasel required\n"

    atf_check -o "inline:${required_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    export ZOGFTW_OPTIONAL_SRC_DATASETS="tank/home/zogftw/tralala"
    optional_dataset_info="tank/home/zogftw/tralala optional\n"
    atf_check -o "inline:${required_dataset_info}${optional_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    export ZOGFTW_EXTERNAL_SRC_DATASETS="tank/home/zogftw/tirili tank/home/zogftw/tirila"
    external_dataset_info="tank/home/zogftw/tirila external\ntank/home/zogftw/tirili external\n"
    atf_check -o "inline:${required_dataset_info}${external_dataset_info}${optional_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    unset ZOGFTW_OPTIONAL_SRC_DATASETS
    atf_check -o "inline:${required_dataset_info}${external_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    unset ZOGFTW_REQUIRED_SRC_DATASETS
    atf_check -o "inline:${external_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information

    unset ZOGFTW_EXTERNAL_SRC_DATASETS
    atf_check -o "empty" -e 'empty' -s 'exit:0' \
        "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information
}

zogftw_get_sorted_src_dataset_information_with_spaces_body() {
    export ZOGFTW_REQUIRED_SRC_DATASETS='tank/home/zogftw/blafasel tank/\ spaces\ /\ in\ /\ ma\ /dataset'
    required_dataset_info="tank/%20spaces%20/%20in%20/%20ma%20/dataset required\ntank/home/zogftw/blafasel required\n"

    atf_check -o "inline:${required_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_sorted_src_dataset_information
}

zogftw_get_src_dataset_information_body() {
    local \
        required_dataset_info \
        optional_dataset_info \
        exteral_dataset_info

    atf_check -o "empty" -e 'empty' -s 'exit:0' \
        "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    export ZOGFTW_REQUIRED_SRC_DATASETS="tank/home/zogftw tank/home/zogftw/blafasel"
    required_dataset_info="tank/home/zogftw required\ntank/home/zogftw/blafasel required\n"

    atf_check -o "inline:${required_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    export ZOGFTW_OPTIONAL_SRC_DATASETS="tank/home/zogftw/tralala"
    optional_dataset_info="tank/home/zogftw/tralala optional\n"
    atf_check -o "inline:${required_dataset_info}${optional_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    export ZOGFTW_EXTERNAL_SRC_DATASETS="tank/home/zogftw/tirili tank/home/zogftw/tirila"
    external_dataset_info="tank/home/zogftw/tirili external\ntank/home/zogftw/tirila external\n"
    atf_check -o "inline:${required_dataset_info}${optional_dataset_info}${external_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    unset ZOGFTW_OPTIONAL_SRC_DATASETS
    atf_check -o "inline:${required_dataset_info}${external_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    unset ZOGFTW_REQUIRED_SRC_DATASETS
    atf_check -o "inline:${external_dataset_info}" \
        -e 'empty' -s 'exit:0' "${ZOGFTW}" cmd zogftw_get_src_dataset_information

    unset ZOGFTW_EXTERNAL_SRC_DATASETS
    atf_check -o "empty" -e 'empty' -s 'exit:0' \
        "${ZOGFTW}" cmd zogftw_get_src_dataset_information
}

zogftw_get_sync_worthy_zpools_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 \
        -o empty -e empty "${ZOGFTW}" cmd zogftw_get_sync_worthy_zpools

    export FAKED_ZPOOLS="testpool"
    atf_check -s exit:0 \
        -o "inline:testpool\n" -e empty "${ZOGFTW}" cmd zogftw_get_sync_worthy_zpools

    export FAKED_ZPOOLS="testpool notsyncworthy"
    atf_check -s exit:0 \
        -o "inline:testpool\n" -e empty "${ZOGFTW}" cmd zogftw_get_sync_worthy_zpools

    export FAKED_ZPOOLS="sourcepool"
    export ZOGFTW_DEST_POOL_PREFIX="dataset/does/not/exist"
    atf_check -s exit:0 \
        -o empty -e empty "${ZOGFTW}" cmd zogftw_get_sync_worthy_zpools
}

zogftw_unattached_label_exists_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_unattached_label_exists(): No label provided\n" \
        "${ZOGFTW}" cmd zogftw_unattached_label_exists

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_unattached_label_exists nosuchlabel

    export FAKED_LABELS="label/fakedpool"
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_unattached_label_exists fakedpool

    export FAKED_LABELS="label/blafasel label/fakedpool label/foobar"
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_unattached_label_exists fakedpool

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_unattached_label_exists stillnosuchlabel
}

zogftw_get_unattached_labels_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels

    export FAKED_LABELS="label/fakedpool"
    atf_check -s exit:0 -o "inline:fakedpool\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels

    export FAKED_LABELS="label/fakedpool"
    export FAKED_GELIS="label/fakedpool.eli"
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels

    export FAKED_LABELS="label/fakedpool3 label/fakedpool label/fakedpool2"
    export FAKED_GELIS="label/fakedpool.eli"
    atf_check -s exit:0 -o "inline:fakedpool3 fakedpool2\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels

    export FAKED_LABELS="label/fakedpool label/labelwithoutpool"
    export FAKED_GELIS=""
    atf_check -s exit:0 -o "inline:fakedpool labelwithoutpool\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels

    export FAKED_GELIS="label/fakedpool.eli label/labelwithoutpool.eli"
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_get_unattached_labels
}

zogftw_get_zfs_property_value_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_zfs_property_value(): No property provided\n" \
        "${ZOGFTW}" cmd zogftw_get_zfs_property_value

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_zfs_property_value(): No dataset provided\n" \
        "${ZOGFTW}" cmd zogftw_get_zfs_property_value mountpoint

    atf_check -s exit:2 -o empty \
        -e "inline:bad property list: invalid property [...]\n" \
        "${ZOGFTW}" cmd zogftw_get_zfs_property_value invalid-property blafasel

    export FAKED_ZFS_PROPERTY_VALUE=/blafasel
    atf_check -s exit:0 -o "inline:${FAKED_ZFS_PROPERTY_VALUE}\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_zfs_property_value mountpoint blafasel

    export FAKED_ZFS_PROPERTY_VALUE=lz4
    atf_check -s exit:0 -o "inline:${FAKED_ZFS_PROPERTY_VALUE}\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_get_zfs_property_value compression blafasel
}

zogftw_get_zpool_size_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_zpool_size(): No zpool name provided\n" \
        "${ZOGFTW}" cmd zogftw_get_zpool_size

    export FAKED_ZPOOL_SIZE=1M
    atf_check -s exit:0 -o "inline:1\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
    export FAKED_ZPOOL_SIZE=1G
    atf_check -s exit:0 -o "inline:1024\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
    export FAKED_ZPOOL_SIZE=1000G
    atf_check -s exit:0 -o "inline:1024000\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
    export FAKED_ZPOOL_SIZE=1T
    atf_check -s exit:0 -o "inline:1048576\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
    export FAKED_ZPOOL_SIZE=1T
    atf_check -s exit:0 -o "inline:1048576\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
    export FAKED_ZPOOL_SIZE=351T
    atf_check -s exit:0 -o "inline:368050176\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_zpool_size testpool
}

zogftw_help_body() {
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_help" \
        -e empty "${ZOGFTW}" help
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_help-v" \
        -e empty "${ZOGFTW}" help -v
}

zogftw_import_zpool_body() {
    local \
        fake_zpool_name=fakedpool \
        expected_output_file=./expected_output
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:zogftw_import_zpool(): No label for '${fake_zpool_name}' found\n" \
        "${ZOGFTW}" cmd zogftw_import_zpool "${fake_zpool_name}"

    cat >$expected_output_file <<EOF
'${fake_zpool_name}' attached
'${fake_zpool_name}' imported
EOF
    export FAKED_LABELS="label/${fake_zpool_name}"
    atf_check -s exit:0 \
        -o "file:${expected_output_file}" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_import_zpool "${fake_zpool_name}"

    cat >$expected_output_file <<EOF
'${fake_zpool_name}' attached
zpool '${fake_zpool_name}' already exists. Import skipped.
EOF
    export FAKED_ZPOOLS="${fake_zpool_name}"
    atf_check -s exit:0 \
        -o "file:${expected_output_file}" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_import_zpool "${fake_zpool_name}"
}

zogftw_import_success_body() {
    local \
        successful_import_message no_pools_specified_message
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    successful_import_messages="'fakedpool' attached\n'fakedpool' imported\n"
    no_pools_specified_message="No pool name specified. Trying all unattached labels: fakedpool\n"

    export FAKED_LABELS=label/fakedpool
    export FAKED_POOLS=fakedpool
    atf_check -s exit:0 \
        -o "inline:${successful_import_messages}" \
        -e empty \
        "${ZOGFTW}" import fakedpool
    atf_check -s exit:0 \
        -o "inline:${no_pools_specified_message}${successful_import_messages}" \
        -e empty \
        "${ZOGFTW}" import
}

zogftw_import_without_geli_on_label_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_LABELS=label/labelwithoutpool
    export DISAPPEARING_LABEL=1
    atf_check -s exit:1 \
        -o empty \
        -e "inline:geli: Cannot open /dev/label/labelwithoutpool: No such file or directory.\n" \
        "${ZOGFTW}" cmd zogftw_import_zpool labelwithoutpool
}

zogftw_import_without_labels_to_attach_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o "inline:No labels to attach found.\n" \
        -e empty "${ZOGFTW}" import
    atf_check -s exit:1 -o "inline:No labels to attach found.\n" \
        -e empty "${ZOGFTW}" import nonexistantpool
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_import_zpool(): No label for 'nonexistantpool' found\n" \
        "${ZOGFTW}" cmd zogftw_import_zpool nonexistantpool
}

zogftw_import_without_zpool_on_geli_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_LABELS=label/labelwithoutpool

    atf_check -s exit:1 \
        -o "inline:'labelwithoutpool' attached\n" \
        -e "inline:cannot import 'labelwithoutpool': no such pool available\n" \
        "${ZOGFTW}" cmd zogftw_import_zpool labelwithoutpool
}

zogftw_init_body() {
    local \
        expected_output_file="./expected_output"

    export ZOGFTW_CONFIG_FILE=./blafasel/non_existant_file

    cat >$expected_output_file <<EOF
Creating ./blafasel
Creating ./blafasel/geli/metadata-backups
Creating ./blafasel/geli/keyfiles
Creating ./blafasel/geli/passphrases
Creating ./blafasel/non_existant_file
Please verify that the locations make sense and that the permissions match your expectations.
EOF
    atf_check -s exit:0 \
        -o "file:${expected_output_file}" -e empty \
        "${ZOGFTW}" init

    cat >$expected_output_file <<EOF
./blafasel already exists
./blafasel/geli/metadata-backups already exists
./blafasel/geli/keyfiles already exists
./blafasel/geli/passphrases already exists
./blafasel/non_existant_file already exists
Please verify that the locations make sense and that the permissions match your expectations.
EOF
    atf_check -s exit:0 \
        -o "file:${expected_output_file}" -e empty \
        "${ZOGFTW}" init

    # It could be argued that this behaviour is a bit silly ...
    export ZOGFTW_CONFIG_FILE=""
    export ZOGFTW_CONFIG_DIRECTORY=""
    export ZOGFTW_GELI_METADATA_BACKUP_DIR=""
    export ZOGFTW_GELI_KEYFILE_DIR=""
    export ZOGFTW_GELI_PASSPHRASE_DIR=""
    atf_check -s exit:0 \
        -o "inline:Please verify that the locations make sense and that the permissions match your expectations.\n" \
        -e empty \
        "${ZOGFTW}" init
}

# While this test case's coverage is incomplete, the initialization
# is also tested by zogftw_get_config_variables_body()'s test case
zogftw_initialize_unset_optional_config_variables_body() {
    source_zogftw

    if [ "${ZOGFTW_USE_GPG_AGENT}" != 0 ]; then
        atf_fail "ZOGFTW_USE_GPG_AGENT has an unexpected default"
    fi
    if [ "${ZOGFTW_BACKUP_DATASET_COMPRESSION}" != "lz4" ]; then
        atf_fail "ZOGFTW_BACKUP_DATASET_COMPRESSION has an unexpected default"
    fi

    export ZOGFTW_USE_GPG_AGENT=1
    export ZOGFTW_BACKUP_DATASET_COMPRESSION=gzip

    zogftw_initialize_unset_optional_config_variables || atf_fail "Failed to reinitialize unset variables"

    if [ "${ZOGFTW_USE_GPG_AGENT}" != 1 ]; then
        atf_fail "ZOGFTW_USE_GPG_AGENT was modified by zogftw_initialize_unset_optional_config_variables"
    fi
    if [ "${ZOGFTW_BACKUP_DATASET_COMPRESSION}" != "gzip" ]; then
        atf_fail "ZOGFTW_BACKUP_DATASET_COMPRESSION was modified by zogftw_initialize_unset_optional_config_variables"
    fi
}

zogftw_invalid_subcommand_body() {
    local \
        error_message

    error_message="$(cat $(atf_get_srcdir)/expected_output/zogftw_help)\nInvalid or missing parameters!\n"
    atf_check -s exit:1 \
        -o empty \
        -e "inline:${error_message}" "${ZOGFTW}" invalid
}

zogftw_label_device_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_label_device(): No label provided\n" \
        "${ZOGFTW}" cmd zogftw_label_device

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_label_device(): No device to label provided\n" \
        "${ZOGFTW}" cmd zogftw_label_device newpool

    atf_check -s exit:1 -o empty \
        -e "inline:Label 's p a c e s' contains spaces which is currently unsupported\n" \
        "${ZOGFTW}" cmd zogftw_label_device "s p a c e s" /dev/null

    atf_check -s exit:0 -o "inline:Labeling /dev/null newpool\n" \
        -e empty "${ZOGFTW}" cmd zogftw_label_device newpool /dev/null
}

zogftw_load_config_file_body() {
    local \
        output expected_output
    source_zogftw

    ZOGFTW_CONFIG_FILE=./non_existant_file
    expected_output="Config file './non_existant_file' does not exist. Use 'zogftw init' to create it."
    output=$(zogftw_load_config_file 2>&1)
    if [ $? != 1 ]; then
        atf_fail "Unexpected return code"
    fi
    if [ "${output}" !=  "${expected_output}" ]; then
        atf_fail "Unexpected output: $output"
    fi

    ZOGFTW_CONFIG_FILE=./zogftw.config
    echo "bla=fasel" > "${ZOGFTW_CONFIG_FILE}"
    output=$(zogftw_load_config_file 2>&1; eval 'echo bla=$bla')
    if [ $? != 0 ]; then
        atf_fail "Unexpected return code"
    fi
    if [ "${output}" != "bla=fasel" ]; then
        atf_fail "Config file wasn't properly sourced: $output"
    fi
    if [ "${ZOGFTW_CONFIG_DIRECTORY}" != "" ]; then
        atf_fail "\$ZOGFTW_CONFIG_DIRECTORY set to '${ZOGFTW_CONFIG_DIRECTORY}'"
    fi
}

# zogftw_main() is tested through most of the other test cases
# already, therefore this one doesn't have to cover it completely.
zogftw_main_body() {
    local \
        expected_stderr_file=./expected_stderr

    cat "$(atf_get_srcdir)/expected_output/zogftw_help" > $expected_stderr_file
    echo "Invalid or missing parameters!" >> $expected_stderr_file

    atf_check -s exit:1 -o empty -e "file:${expected_stderr_file}" \
        "${ZOGFTW}" cmd zogftw_main

    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_main nop
}

zogftw_nop_body() {
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" nop
}

zogftw_normalize_dataset_body() {
    local \
        test_directory="bla/fasel"

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    mkdir -p "${test_directory}" || atf_fail "Failed to create test directory ${test_directory}"
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_normalize_dataset(): ${test_directory} is neither a valid dataset name nor an existing directory you have access to\n" \
        "${ZOGFTW}" cmd zogftw_normalize_dataset "${test_directory}"

    export FAKED_DATASET=${test_directory}
    atf_check -s exit:0 -o "inline:${test_directory}\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_normalize_dataset "${test_directory}"
}

zogftw_pipe_buffer_body() {
    local \
        piped_message="buffer me softly"
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    echo "${piped_message}" | atf_check -s exit:0 \
        -o "inline:${piped_message}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_pipe_buffer
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_pipe_buffer < /dev/null
}

zogftw_prepare_pool_for_receiving_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 \
        -o "inline:Creating backup structure newpool/backup/testhostname\n" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_prepare_pool_for_receiving newpool
}

zogftw_prime_sudo_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_prime_sudo
}

zogftw_get_receive_resume_token_body() {
    local \
        token_value
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_get_receive_resume_token(): Missing destination dataset\n" \
        "${ZOGFTW}" cmd zogftw_get_receive_resume_token

    atf_check -s exit:1 -o "inline:-\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_receive_resume_token invalid

    token_value="1-f504d3a71-d8-789c636064000310a500c4ec50360710e72765a52697303054f042d460c8a7a515a7968064b8e0f26c48f2499525a9c5403a83e3e843262cfa4bf2d34b3353181856345e5bcd7f95e59207923c27583e2f3137154827e665eb67e4e7a6eaa765eb2796a664e6eb97e5172565163b1819189aea1a1aea1a58c61b1a59199a31c00100cb3e2176"

    export FAKED_ZFS_PROPERTY_VALUE="${token_value}"
    atf_check -s exit:0 -o "inline:${FAKED_ZFS_PROPERTY_VALUE}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_receive_resume_token \
        testpool/backup/testhostname/sourcepool/blafasel

    # Currently the value is not sanity checked.
    export FAKED_ZFS_PROPERTY_VALUE=blafasel
    atf_check -s exit:0 -o "inline:${FAKED_ZFS_PROPERTY_VALUE}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_receive_resume_token \
        testpool/backup/testhostname/sourcepool/tralala
}

zogftw_resume_transfer_body() {
    local \
        token_value dataset

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_resume_transfer(): Missing receive_resume_token\n" \
        "${ZOGFTW}" cmd zogftw_resume_transfer

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_resume_transfer(): Missing destination dataset\n" \
        "${ZOGFTW}" cmd zogftw_resume_transfer -

    token_value="1-f504d3a71-d8-789c636064000310a500c4ec50360710e72765a52697303054f042d460c8a7a515a7968064b8e0f26c48f2499525a9c5403a83e3e843262cfa4bf2d34b3353181856345e5bcd7f95e59207923c27583e2f3137154827e665eb67e4e7a6eaa765eb2796a664e6eb97e5172565163b1819189aea1a1aea1a58c61b1a59199a31c00100cb3e2176"
    dataset="testpool/backup/testhostname/sourcepool/blafasel/die/da"
    cat >expected_stdout <<EOF
Resuming transfer to ${dataset} using token $token_value
Fake sudo: 'zfs received': Fake stream package based on token
EOF
    export FAKED_HOSTNAME="testhostname"
    atf_check -s exit:0 -o "file:expected_stdout" \
        -e "inline:Faking resumed send operation\n" \
        "${ZOGFTW}" cmd zogftw_resume_transfer "${token_value}" "${dataset}"

    cat >expected_stdout <<EOF
Resuming transfer to testpool/backup/testhostname/sourcepool/blafasel/die/da using token invalid-token
Fake sudo: Did not 'zfs receive' anything
EOF
    token_value="invalid-token"
    atf_check -s exit:1 -o "file:expected_stdout" \
        -e "inline:cannot resume send: resume token is corrupt (invalid format)\n" \
        "${ZOGFTW}" cmd zogftw_resume_transfer "${token_value}" "${dataset}"
}

zogftw_auto_resume_body() {
    local \
        token_value dataset

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_auto_resume(): Missing destination dataset\n" \
        "${ZOGFTW}" cmd zogftw_auto_resume

    export FAKED_HOSTNAME="testhostname"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_auto_resume(): No receive_resume_token found on testpool/backup/testhostname/sourcepool/blafasel\n" \
        "${ZOGFTW}" cmd zogftw_auto_resume testpool/backup/testhostname/sourcepool/blafasel

    token_value="1-f504d3a71-d8-789c636064000310a500c4ec50360710e72765a52697303054f042d460c8a7a515a7968064b8e0f26c48f2499525a9c5403a83e3e843262cfa4bf2d34b3353181856345e5bcd7f95e59207923c27583e2f3137154827e665eb67e4e7a6eaa765eb2796a664e6eb97e5172565163b1819189aea1a1aea1a58c61b1a59199a31c00100cb3e2176"
    dataset="testpool/backup/testhostname/sourcepool/blafasel/die/da"

    cat >expected_stdout <<EOF
Resuming transfer to ${dataset} using token ${token_value}
Fake sudo: 'zfs received': Fake stream package based on token
EOF
    export FAKED_ZFS_PROPERTY_VALUE="${token_value}"
    atf_check -s exit:0 -o "file:expected_stdout" \
        -e "inline:Faking resumed send operation\n" \
        "${ZOGFTW}" cmd zogftw_auto_resume "${dataset}"
}

zogftw_receive_resume_token_belongs_to_dataset_body() {
    local \
        dataset token;

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_validate_receive_resume_token(): Missing receive_resume_token\n" \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_validate_receive_resume_token(): Missing source dataset\n" \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "token"

    dataset="sourcepool/dataset/does/not/exist"
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_validate_receive_resume_token(): Dataset $dataset does not exist\n" \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "token" "${dataset}"

    token="flawed-token-1"
    dataset="sourcepool/blafasel/die/da"
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_validate_receive_resume_token(): Failed to get toguid\n" \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "${token}" "${dataset}"

    token="valid-token-1"
    dataset="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_validate_receive_resume_token(): No snapshot sourcepool/blafasel/die/da@2014-08-09_13:20 found on sourcepool/blafasel/die/da\n" \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "${token}" "${dataset}"

    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20 2015-08-09_13:02"

    export FAKED_ZFS_PROPERTY_VALUE="12345678901234567890"
    atf_check -s exit:1 -e empty \
        -o 'inline:zogftw_validate_receive_resume_token(): Found snapshot sourcepool/blafasel/die/da@2014-08-09_13:20 but guid 12345678901234567890 does not match toguid 15135829836139206824\n' \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "${token}" "${dataset}"

    export FAKED_ZFS_PROPERTY_VALUE="15135829836139206824"
    atf_check -s exit:0 -e empty -o empty \
        "${ZOGFTW}" cmd zogftw_receive_resume_token_belongs_to_dataset "${token}" "${dataset}"
}

zogftw_send_dataset_incrementally_body() {
    local \
        source_dataset destination_dataset \
        expected_output_file=./expected_output
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_send_dataset_incrementally(): Missing source dataset\n" \
        "${ZOGFTW}" cmd zogftw_send_dataset_incrementally

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_send_dataset_incrementally(): Missing destination dataset\n" \
        "${ZOGFTW}" cmd zogftw_send_dataset_incrementally blafasel

    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    source_dataset=sourcepool/blafasel/die/da
    destination_dataset="testpool/backup/$(hostname -s)/sourcepool/blafasel/die/da"
    cat >$expected_output_file <<EOF
Destination dataset '${destination_dataset}' doesn't seem to exist yet
Creating testpool/backup/testhostname/sourcepool/blafasel/die as parent for ${destination_dataset}
Fake sudo: 'zfs received': Fake stream package for ${source_dataset}@2014-08-09_13:18
EOF
    atf_check -s exit:0 \
        -o "file:${expected_output_file}" \
        -e empty \
        "${ZOGFTW}" cmd zogftw_send_dataset_incrementally "${source_dataset}" "${destination_dataset}"
}

zogftw_snapshot_body() {
    local \
        nonexistant_dataset=sourcepool/dataset/does/not/exist
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:Dataset '${nonexistant_dataset}' does not exist\n" \
        "${ZOGFTW}" snapshot $nonexistant_dataset
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_snapshot(): No dataset to snapshot provided and no default specified.\n" \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS="${nonexistant_dataset}"
    atf_check -s exit:1 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${nonexistant_dataset}\n" \
        -e "inline:Dataset '${nonexistant_dataset}' does not exist\n" \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS="sourcepool/foo/bar sourcepool/no/sudo/required"
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS="sourcepool/foo/bar sourcepool/no/sudo/required sourcepool/dataset/does/not/exist"
    atf_check -s exit:1 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e "inline:Dataset '${nonexistant_dataset}' does not exist\n" \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS="sourcepool/foo/bar sourcepool/dataset/does/not/exist sourcepool/no/sudo/required"
    atf_check -s exit:1 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e "inline:Dataset '${nonexistant_dataset}' does not exist\n" \
        "${ZOGFTW}" snapshot

    for snapshot_cmd in snapshot snap; do
        atf_check -s exit:0 -o empty -e empty \
            "${ZOGFTW}" "${snapshot_cmd}" sourcepool/foo/bar
        atf_check -s exit:0 -o empty -e empty \
            "${ZOGFTW}" "${snapshot_cmd}" sourcepool/no/sudo/required
        atf_check -s exit:0 -o empty -e empty \
            "${ZOGFTW}" "${snapshot_cmd}" "sourcepool/ spaces are / awesome/ "
    done
}

zogftw_snapshot_dataset_body() {
    local \
        nonexistant_dataset=sourcepool/dataset/does/not/exist
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_snapshot_dataset(): No dataset to snapshot provided.\n" \
        "${ZOGFTW}" cmd zogftw_snapshot_dataset
    atf_check -s exit:1 -o empty \
        -e "inline:Dataset '${nonexistant_dataset}' does not exist\n" \
        "${ZOGFTW}" cmd zogftw_snapshot_dataset $nonexistant_dataset

    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_snapshot_dataset sourcepool/foo/bar
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_snapshot_dataset sourcepool/no/sudo/required
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_snapshot_dataset "sourcepool/ spaces are / awesome/ "
}

zogftw_get_datasets_recursively_body() {

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZFS_DATASETS="tank
tank/bla
tank/bla/fasel"

    atf_check -s exit:0 -o "inline:${FAKED_ZFS_DATASETS}\n" -e empty \
        "${ZOGFTW}" cmd zogftw_get_datasets_recursively "tank"
}

zogftw_restore_failures_body() {

    atf_check -s exit:1 -o empty -e 'inline:No destination zpool specified\n' \
        "${ZOGFTW}" restore

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:No imported zpools managed by zogftw detected\n" \
        "${ZOGFTW}" restore tank
}

# Create a local zogftw copy that doesn't try to prevent
# the passphrase from being echoed to a terminal and thus
# doesn't mind running in the environment setup by atf-check.
create_stty_less_zogftw_copy() {
    grep -v stty "${ZOGFTW}" > zogftw || atf_fail "Failed to create stty-free zogftw copy"
    chmod +x zogftw
}

zogftw_request_passphrase_body() {
    local \
        expected_prompt expected_passphrase passphrase

    expected_prompt="Passphrase:"
    expected_passphrase="My secret passphrase"

    echo "${expected_passphrase}" | atf_check -s exit:1 \
        -o "inline:${expected_prompt}" -e "inline:stty: stdin isn't a terminal\n" \
        "${ZOGFTW}" cmd zogftw_request_passphrase "${expected_prompt}" "passphrase"

    create_stty_less_zogftw_copy

    echo "${expected_passphrase}" | atf_check -s exit:0 \
        -o "inline:${expected_prompt}\n" -e "empty" \
        sh ./zogftw cmd zogftw_request_passphrase "${expected_prompt}" "passphrase"
    echo "${expected_passphrase}" | atf_check -s exit:0 \
        -o "inline:${expected_prompt}\n" -e "inline:${expected_passphrase}\n" \
        sh ./zogftw cmd eval 'zogftw_request_passphrase "'${expected_prompt}'" "passphrase" ; echo $passphrase 1>&2'
}

zogftw_replace_pattern_body() {
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_replace: No string provided\n" \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" ""
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a\ b\ c\ d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a\ b\ c%20d"
    atf_check -s exit:0 -o "inline:a%20b%20c d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a%20b\ c d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a\ b\ c\ \ d"
    atf_check -s exit:0 -o "inline:a%20b%20c%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a\ b\ c\ %20d"
    atf_check -s exit:0 -o "inline:a%20b%20%20c%20%20%20d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "a\ b\ \ c\ \ \ d"
    atf_check -s exit:0 -o "inline:%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "\ "
    atf_check -s exit:0 -o "inline:%20%20%20%20%20%20\n" -e empty \
        "${ZOGFTW}" cmd zogftw_replace_pattern "\ " "%20" "\ \ \ \ \ \ "
}

zogftw_show_hook_capabilities_body() {
    local \
        main_hook_output

    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" nop

    export ZOGFTW_SHOW_AVAILABLE_HOOKS=1
    main_hook_output="zogftw_main_hook() has access to: mode: 'nop', param_1: '', param_2: ''\n"
    main_hook_output="$main_hook_output"'zogftw_fyi_message_emitted_hook() has access to the variables $information $zogftw_fyi_prefix and $log_message\n'
    atf_check -s exit:0 -o "inline:${main_hook_output}" -e empty "${ZOGFTW}" nop
}

zogftw_snapshot_creation_failures_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:Dataset 'nonexistantpool/bla/fasel' does not exist\n" \
        "${ZOGFTW}" snapshot nonexistantpool/bla/fasel
}

zogftw_snapshot_creation_using_defaults_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='sourcepool/no/sudo/required'
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty "${ZOGFTW}" snap

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='sourcepool/no/sudo/required sourcepool/foo/bar'
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='sourcepool/\ spaces\ are\ /\ awesome/\ '
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='sourcepool/no/sudo/required sourcepool/\ spaces\ are\ /\ awesome/\ '
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS="sourcepool/no/sudo/required
                                             sourcepool/\ spaces\ are\ /\ awesome/\ 
                                             sourcepool/no/sudo/required"
    atf_check -s exit:0 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e empty \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='sourcepool/no/sudo/required nonexistantpool/bla/fasel'
    atf_check -s exit:1 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e "inline:Dataset 'nonexistantpool/bla/fasel' does not exist\n" \
        "${ZOGFTW}" snapshot

    export ZOGFTW_DEFAULT_SNAPSHOT_DATASETS='nonexistantpool/bla/fasel sourcepool/no/sudo/required'
    atf_check -s exit:1 \
        -o "inline:No dataset to snapshot specified. Snapshotting: ${ZOGFTW_DEFAULT_SNAPSHOT_DATASETS}\n" \
        -e "inline:Dataset 'nonexistantpool/bla/fasel' does not exist\n" \
        "${ZOGFTW}" snapshot
}

zogftw_sudo_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_sudo false
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_sudo glabel label newpool /dev/null
}

zogftw_successful_snapshot_creation_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" snapshot sourcepool/foo/bar
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" snap sourcepool/foo/bar
}

zogftw_successful_snapshot_creation_without_sudo_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" snap sourcepool/no/sudo/required
}

zogftw_sync_creating_destination_dataset_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    cat > ./expected_output_1 <<EOF
Destination dataset 'testpool/backup/testhostname/sourcepool/blafasel/die/da' doesn't seem to exist yet
Creating testpool/backup/testhostname/sourcepool/blafasel/die as parent for testpool/backup/testhostname/sourcepool/blafasel/die/da
Fake sudo: 'zfs received': Fake stream package for sourcepool/blafasel/die/da@2014-08-09_13:18
EOF
    atf_check -s exit:0 -o "file:./expected_output_1" -e empty "${ZOGFTW}" sync testpool

    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    cat > ./expected_output_2 <<EOF
Destination dataset 'testpool/backup/testhostname/sourcepool/blafasel/die/da' doesn't seem to exist yet
Creating testpool/backup/testhostname/sourcepool/blafasel/die as parent for testpool/backup/testhostname/sourcepool/blafasel/die/da
Fake sudo: 'zfs received': Fake stream package for sourcepool/blafasel/die/da@2014-08-09_13:18
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output_2" -e empty "${ZOGFTW}" sync testpool

    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19"
    cat > ./expected_output_3 <<EOF
Destination dataset 'testpool/backup/testhostname/sourcepool/blafasel/die/da' doesn't seem to exist yet
Creating testpool/backup/testhostname/sourcepool/blafasel/die as parent for testpool/backup/testhostname/sourcepool/blafasel/die/da
Fake sudo: 'zfs received': Fake stream package for sourcepool/blafasel/die/da@2014-08-09_13:18
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:19. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output_3" -e empty "${ZOGFTW}" sync testpool
}

zogftw_sync_skip_required_dataset_creation_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_ZFS_PROPERTY_VALUE="no"

    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" sync testpool
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_sync_zpool testpool

    export ZOGFTW_LOG_SKIPPED_DATASETS="1"
    cat > ./expected_output <<EOF
Skipping required src dataset '${ZOGFTW_REQUIRED_SRC_DATASETS}': pool '${FAKED_ZPOOLS}' does not get new datasets.
EOF
    atf_check -s exit:0 -o "file:./expected_output" -e empty "${ZOGFTW}" sync testpool
    atf_check -s exit:0 -o "file:./expected_output" -e empty "${ZOGFTW}" cmd zogftw_sync_zpool testpool

    cat > ./expected_output <<EOF
No destination pool specified. Synchronizing all pools prepared for receiving: testpool
Skipping required src dataset '${ZOGFTW_REQUIRED_SRC_DATASETS}': pool '${FAKED_ZPOOLS}' does not get new datasets.
EOF
    atf_check -s exit:0 -o "file:./expected_output" -e empty "${ZOGFTW}" sync

    # XXX: We're currently don't verify that an existing dataset without
    #      snapshots (a placeholder dataset) does not get new ones.
}

zogftw_sync_failures_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:No destination pool specified and no prepared pools detected\n" "${ZOGFTW}" sync
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_sync_zpool(): Destination pool 'nonexistantpool' does not exist\n" "${ZOGFTW}" sync nonexistantpool

    export ZOGFTW_REQUIRED_SRC_DATASETS="nonexistantpool/blafasel"
    atf_check -s exit:1 -o empty -e "inline:Can't find required source dataset 'nonexistantpool/blafasel'\n" "${ZOGFTW}" sync testpool
}

zogftw_sync_with_existing_destination_dataset_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    cat > ./expected_output <<EOF
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output" \
        -e "ignore" "${ZOGFTW}" sync testpool
}

zogftw_sync_zpool_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_sync_zpool(): No destination pool specified\n" \
        "${ZOGFTW}" cmd zogftw_sync_zpool

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_sync_zpool(): Destination pool 'fakedpool' does not exist\n" \
        "${ZOGFTW}" cmd zogftw_sync_zpool fakedpool

    # This is a variation of the previous test case
    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    cat > ./expected_output <<EOF
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output" \
        -e empty "${ZOGFTW}" cmd zogftw_sync_zpool testpool
}

zogftw_sync_resume_followed_by_vanilly_receive_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1

    token_value="valid-token-1"
    dataset="testpool/backup/testhostname/sourcepool/blafasel/die/da"
    export FAKED_ZFS_PROPERTY_VALUE="${token_value}"
    export FAKED_ZFS_GUID_VALUE="15135829836139206824"

    cat > ./expected_output <<EOF
Resuming transfer to testpool/backup/testhostname/sourcepool/blafasel/die/da using token ${token_value}
Fake sudo: 'zfs received': Fake zfs send stream resumed using ${token_value}
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output" \
        -e empty "${ZOGFTW}" cmd zogftw_sync_zpool testpool
}

zogftw_sync_with_unsyncable_destination_dataset_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2013-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    cat > ./expected_output_4 <<EOF
WARNING: could not send sourcepool/blafasel/die/da@2014-08-09_13:18:
incremental source (sourcepool/blafasel/die/da@2013-08-09_13:18) does not exist
EOF
    atf_check -s exit:1 -o "inline:Fake sudo: Did not 'zfs receive' anything\n" \
        -e "file:./expected_output_4" "${ZOGFTW}" sync testpool
}

zogftw_sync_without_src_datasets_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZPOOLS="testpool"
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" sync testpool
}

zogftw_transfer_first_snapshot_body() {
    local \
        destination_dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_transfer_first_snapshot(): Missing source dataset\n" \
        "${ZOGFTW}" cmd zogftw_transfer_first_snapshot
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_transfer_first_snapshot(): Missing destination dataset\n" \
        "${ZOGFTW}" cmd zogftw_transfer_first_snapshot blafasel
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_transfer_first_snapshot(): Source dataset 'sourcepool/dataset/does/not/exist' does not exist\n" \
        "${ZOGFTW}" cmd zogftw_transfer_first_snapshot sourcepool/dataset/does/not/exist blafasel

    # This is a variation of test case zogftw_sync_with_existing_destination_dataset
    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/blafasel/die/da"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    destination_dataset="testpool/backup/$(hostname -s)/sourcepool/blafasel/die/da"
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_transfer_first_snapshot(): Destination dataset '${destination_dataset}' already exists and contains snapshots\n" \
        "${ZOGFTW}" cmd zogftw_transfer_first_snapshot sourcepool/blafasel/die/da "${destination_dataset}"
}

zogftw_transfer_first_snapshot_with_spaces_body() {
    local \
        destination_dataset
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZPOOLS="testpool"
    export ZOGFTW_REQUIRED_SRC_DATASETS="sourcepool/ spaces are / awesome/ "
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    cat > ./expected_output <<EOF
Creating testpool/backup/testhostname/sourcepool/ spaces are / awesome as parent for testpool/backup/testhostname/sourcepool/ spaces are / awesome/ 
Fake sudo: 'zfs received': Fake stream package for sourcepool/ spaces are / awesome/ @2014-08-09_13:18
EOF
    destination_dataset="testpool/backup/$(hostname -s)/sourcepool/ spaces are / awesome/ "
    atf_check -s exit:0 -o "file:./expected_output" -e empty \
        "${ZOGFTW}" cmd zogftw_transfer_first_snapshot "${ZOGFTW_REQUIRED_SRC_DATASETS}" "${destination_dataset}"
}

zogftw_transfer_last_snapshot_body() {
    local \
        incremental_mode
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export ZOGFTW_STORE_LAST_SNAPSHOT_AS_PROPERTY=0

    incremental_mode="-F"
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_transfer_last_snapshot(): Invalid incremental mode: ${incremental_mode}\n" \
        "${ZOGFTW}" cmd zogftw_transfer_last_snapshot $incremental_mode

    for incremental_mode in "-i" "-I"; do
        atf_check -s exit:1 -o empty \
            -e "inline:zogftw_transfer_last_snapshot(): Missing source dataset\n" \
            "${ZOGFTW}" cmd zogftw_transfer_last_snapshot "${incremental_mode}"
        atf_check -s exit:1 -o empty \
            -e "inline:zogftw_transfer_last_snapshot(): Missing destination dataset\n" \
            "${ZOGFTW}" cmd zogftw_transfer_last_snapshot "${incremental_mode}" blafasel
        atf_check -s exit:1 -o empty \
            -e "inline:zogftw_transfer_last_snapshot(): Source dataset 'sourcepool/dataset/does/not/exist' does not exist\n" \
            "${ZOGFTW}" cmd zogftw_transfer_last_snapshot "${incremental_mode}" sourcepool/dataset/does/not/exist blafasel
    done

    # This is a variation of the test case: zogftw_sync_with_existing_destination_dataset
    export FAKED_ZPOOLS="testpool"
    export FAKED_SOURCE_SNAPSHOT_NAMES="2014-08-09_13:18 2014-08-09_13:19 2014-08-09_13:20"
    export FAKED_DESTINATION_SNAPSHOT_NAMES="2014-08-09_13:18"
    export FAKED_DESTINATION_DATASET_EXISTS=1
    cat > ./expected_output <<EOF
Fake sudo: 'zfs received': Fake incremental stream package for sourcepool/blafasel/die/da. Start: @2014-08-09_13:18. End: @2014-08-09_13:20
EOF
    atf_check -s exit:0 -o "file:./expected_output" \
        -e empty "${ZOGFTW}" cmd zogftw_transfer_last_snapshot -I sourcepool/blafasel/die/da \
        "testpool/backup/$(hostname -s)/sourcepool/blafasel/die/da"
}

zogftw_unescape_spaces_body() {
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_unescape_spaces: No string provided\n" \
        "${ZOGFTW}" cmd zogftw_unescape_spaces ""
    atf_check -s exit:0 -o "inline:a b c d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a b c d"
    atf_check -s exit:0 -o "inline:a b c d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a%20b%20c%20d"
    atf_check -s exit:0 -o "inline:a b c d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a%20b%20c d"
    atf_check -s exit:0 -o "inline:a b c  d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a%20b%20c  d"
    atf_check -s exit:0 -o "inline:a b c  d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a%20b%20c%20%20d"
    atf_check -s exit:0 -o "inline:a b c  d\n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "a b c%20%20d"
    atf_check -s exit:0 -o "inline: \n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "%20"
    atf_check -s exit:0 -o "inline:      \n" -e empty \
        "${ZOGFTW}" cmd zogftw_unescape_spaces "%20%20%20%20%20%20"
}

zogftw_update_last_snapshot_properties_body() {
    atf_check -s exit:1 -o empty -e 'inline:No zpool to update specified\n' \
        "${ZOGFTW}" update-last-snapshot-properties

    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 \
        -o empty \
        -e "inline:'unitialisedpool/backup/testhostname' doesn't exist. Nothing to update.\n" \
        "${ZOGFTW}" update-last-snapshot-properties unitialisedpool

    cat > ./expected_output <<EOF
Updating de.fabiankeil:zogftw:last_snapshot properties for pooltoupdate ...
pooltoupdate/backup/testhostname contains no snapshots. Skipping it.
pooltoupdate/backup/testhostname/bla contains no snapshots. Skipping it.
Setting last snapshot property value pooltoupdate/backup/testhostname/bla/fasel@2022-03-30_15:36
EOF
    atf_check -s exit:0 \
        -o file:./expected_output \
        -e empty \
        "${ZOGFTW}" update-last-snapshot-properties pooltoupdate
}

zogftw_zpool_gets_new_datasets_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_zpool_gets_new_datasets(): No pool specified\n" \
        "${ZOGFTW}" cmd zogftw_zpool_gets_new_datasets

    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_zpool_gets_new_datasets(): 'nonexistantpool/backup/testhostname' doesn't exist on pool 'nonexistantpool'\n" \
        "${ZOGFTW}" cmd zogftw_zpool_gets_new_datasets nonexistantpool

    export FAKED_ZFS_PROPERTY_VALUE="yes"
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_zpool_gets_new_datasets testpool

    export FAKED_ZFS_PROPERTY_VALUE="no"
    atf_check -s exit:1 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_zpool_gets_new_datasets testpool

    export FAKED_ZFS_PROPERTY_VALUE="-"
    atf_check -s exit:0 -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_zpool_gets_new_datasets testpool

}

zogftw_user_has_zfs_permission_body() {
    local \
        permission expected_exit_code
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    for permission in create diff mount receive send snapshot; do
        atf_check -s exit:0 -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_user_has_zfs_permission $permission \
            sourcepool/no/sudo/required
    done

    if [ "$(id -u)" != 0 ]; then
        expected_exit_code=1
    else
        expected_exit_code=0
    fi
    atf_check -s exit:$expected_exit_code -o empty -e empty \
        "${ZOGFTW}" cmd zogftw_user_has_zfs_permission snapshot sourcepool/foo/bar

    for permission in pet walk desire; do
        atf_check -s exit:$expected_exit_code -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_user_has_zfs_permission $permission \
            sourcepool/no/sudo/required
        atf_check -s exit:$expected_exit_code -o empty -e empty \
            "${ZOGFTW}" cmd zogftw_user_has_zfs_permission $permission \
            sourcepool/foo/bar
    done
}

zogftw_usage_body() {
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_help" \
        -e empty "${ZOGFTW}" cmd zogftw_usage
    atf_check -s exit:0 \
        -o "file:$(atf_get_srcdir)/expected_output/zogftw_help-v" \
        -e empty "${ZOGFTW}" cmd zogftw_usage -v
}

zogftw_wtf_body() {
    atf_check -s exit:1 -o empty -e "inline:Blafasel\n" "${ZOGFTW}" cmd zogftw_wtf "Blafasel"
    atf_check -s exit:1 -o empty -e "inline:zogftw_wtf(): No complaints?\n" "${ZOGFTW}" cmd zogftw_wtf ""
}

zogftw_zpool_does_exist_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    atf_check -s exit:1 -o empty \
        -e 'inline:zogftw_zpool_does_exist(): No zpool provided\n' \
        "${ZOGFTW}" cmd zogftw_zpool_does_exist
    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_does_exist nonexistantpool
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_does_exist testpool
}

zogftw_zpool_is_puny_body() {
    enable_pseudo_executables || atf_fail "Failed to enable pseudo executables"

    export FAKED_ZPOOL_SIZE=1073741824 # 1G
    atf_check -s exit:0 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_is_puny testpool
    export FAKED_ZPOOL_SIZE=4294967296 # 4G, the ZOGFTW_MAX_SPACE_CONSTRAINED_ZPOOL_SIZE default
    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_is_puny testpool
    export FAKED_ZPOOL_SIZE=107374182400 # 100G
    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_is_puny testpool
    export FAKED_ZPOOL_SIZE=1099511627776 # 1T
    atf_check -s exit:1 -o empty -e empty "${ZOGFTW}" cmd zogftw_zpool_is_puny testpool
    export FAKED_ZPOOL_SIZE=NaN
    atf_check -s exit:1 -o empty \
        -e "inline:zogftw_zpool_is_puny(): Can't figure out the zpool size for 'testpool'\n" \
        "${ZOGFTW}" cmd zogftw_zpool_is_puny testpool
}
